from __future__ import annotations

import anyio
import time
import os
import sys
from contextlib import asynccontextmanager, nullcontext
from tempfile import TemporaryDirectory
from pathlib import Path as FSPath
from functools import partial

from moat.link.client import Link
from moat.link.server import Server
from moat.link.backend import get_backend
from moat.util import (  # pylint:disable=no-name-in-module
    CFG,  # noqa:F401
    ensure_cfg,
    CtxObj,
    attrdict,
    combine_dict,
    NotGiven,
    Path,P,
    Root,
    ValueEvent,
)

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, AsyncIterator,Self

ensure_cfg("moat.link")


otm = time.time

_seq = 0


# TODO launch a "real" broker instead


async def run_broker(cfg, *, task_status):
    """
    Runs a basic MQTT broker.

    The task status returns the port we're listening on.
    """
    #cfg  # noqa:B018  # pyright:ignore
    #broker = AsyncMQTTBroker(("127.0.0.1", 0))
    #await broker.serve(task_status=task_status)
    port=40000+(os.getpid()+123)%10000
    async with (
            anyio.NamedTemporaryFile(mode="w+") as tf,
            anyio.create_task_group() as tg,
            ):
        await tf.write(f"""\
allow_anonymous true
retained_messages_mode enabled_without_persistence
thread_count 1

listen {{
    protocol mqtt
    port { port }
    inet4_bind_address 127.0.0.1
}}
""")
        await tf.flush()

        tg.start_soon(partial(anyio.run_process,["flashmq","-c",tf.name],stderr=sys.stderr,stdout=sys.stdout))
        for _ in range(20):
            try:
                sock = await anyio.connect_tcp("127.0.0.1",port)
            except EnvironmentError:
                await anyio.sleep(0.1)
            else:
                await sock.aclose()
                break
        else:
            raise RuntimeError("Could not connect to FlashMQ")

        task_status.started(port)


class Scaffold(CtxObj):
    """
    Basic testcase runner for testing with an ephemeral MQTT server.
    """
    tempdir:str|None

    def __init__(self, cfg: attrdict, use_servers=True, tempdir: str | None = None):
        ensure_cfg("moat.link.server", cfg)
        self.cfg = cfg.link

        self.cfg.setdefault("backend", attrdict())
        self.cfg.backend.setdefault("driver", "mqtt")
        self.cfg.backend.setdefault("codec", "std-cbor")
        self.cfg.backend.setdefault("keep_alive", 9999)

        self.cfg.server.ping.cycle=.5
        self.cfg.server.ping.gap=.15
        self.cfg.server.ping.override=True
        self.cfg.server.timeout.startup=3

        self._tempdir = tempdir

        if not use_servers:
            self.cfg.client.init_timeout = None

    @asynccontextmanager
    async def _ctx(self) -> AsyncIterator[Self]:
        Root.set(self.cfg.root)

        with (
            nullcontext(
                self._tempdir,
            )
            if self._tempdir is not None
            else TemporaryDirectory() as tempdir
        ):
            self.tempdir = FSPath(tempdir)
            async with anyio.create_task_group() as self.tg:
                bport = await self.tg.start(run_broker, self.cfg)

                self.cfg.backend.port = bport
                yield self
                self.tg.cancel_scope.cancel()  # pyright:ignore

    async def backend(self, cfg: dict | None = None, **kw):
        """
        Start a backend (background task)
        """
        return await self.tg.start(self._run_backend, cfg, kw)

    @asynccontextmanager
    async def backend_(self, cfg: dict | None, **kw) -> Backend:
        """
        Start a backend (async context manager).
        """
        cfg = combine_dict(cfg, self.cfg, cls=attrdict) if cfg else self.cfg
        async with get_backend(cfg, **kw) as bk:
            yield bk

    async def _run_backend(self, cfg: dict | None, kw: dict, *, task_status) -> Backend:
        """
        Start a backend (Helper).
        """
        async with self.backend_(cfg, **kw) as bk:
            task_status.started(bk)
            await anyio.sleep_forever()
            assert False  # noqa:B011,PT015

    async def server(self, cfg: dict | None = None, **kw) -> tuple[Server, list[dict]]:
        """
        Start a server (background task)

        Returns the server object and the ports it runs on.
        """
        return await self.tg.start(self._run_server, cfg, kw)

    async def _run_server(self, cfg, kw, *, task_status) -> None:
        """
        Run a basic MoaT-Link server. (Helper task)
        """
        cfg = combine_dict(cfg, self.cfg, cls=attrdict) if cfg else self.cfg
        if "ports" in cfg["server"]:
            cfg["server"]["ports"]["main"]["port"] = 0
        cfg["server"]["port"] = 0
        if self.tempdir is not None:
            cfg["server"]["save"]["dir"] = self.tempdir / "data"
        global _seq  # noqa:PLW0603
        _seq += 1

        s = Server(cfg, f"S_{_seq}", **kw)
        await s.serve(task_status=task_status)

    @asynccontextmanager
    async def server_(self, cfg:dict|None=None, **kw) -> None:
        """
        Runs a basic MoaT-Link server. (async context manager)
        """
        async with anyio.create_task_group() as tg:
            yield await tg.start(self._run_server,cfg,kw)
            tg.cancel_scope.cancel()

    async def client(self, *a, **kw):
        """
        Start a client (background task)
        """
        cl = await self.tg.start(partial(self._run_client, *a, **kw))
        return cl

    async def _run_client(self, *a, task_status, **kw) -> Never:
        async with self.client_(*a, **kw) as li:
            task_status.started(li)
            await anyio.sleep_forever()
            assert False  # noqa:B011,PT015

    @asynccontextmanager
    async def client_(self, cfg:dict|None=None, cli:LinkCommon|None=None, name=None) -> Never:
        """
        Start a client (async context manager)
        """
        if cli is None:
            cfg = combine_dict(cfg, self.cfg, cls=attrdict) if cfg else self.cfg

            global _seq  # noqa:PLW0603
            _seq += 1
            name = f"C_{_seq}"

            cli = Link(cfg,name)

        async with cli as li:
            yield li

    @asynccontextmanager
    async def do_watch(self, path, exp=NotGiven, n=0, *a, **kw):
        """
        Run a client that expects @xp and appends all non-exp
        results to @rd.

        All other args+kw are forwarded to `Link.d_watch`.
        """
        async with anyio.create_task_group() as tg, self.client_() as c:
            evt = ValueEvent()
            got=0
            @tg.start_soon
            async def work():
                res=[]
                try:
                    async with c.d_watch(path,*a,**kw) as mon:
                        async for r in mon:
                            if not kw.get("meta") and not kw.get("subtree"):
                                r = (r,)
                            if kw.get("meta"):
                                t = time.time()
                                assert t - 1 < r[-1].timestamp < t
                            if exp is not NotGiven and (not kw.get("subtree") or not len(r[0])) and r[kw.get("subtree",0)] == exp:
                                return
                            res.append(r)
                            if len(res)==n:
                                return
                finally:
                    evt.set(res)
            yield evt
            tg.cancel_scope.cancel()






"""
This module contains the basic MoaT-Link data model.
"""

from __future__ import annotations

from logging import getLogger

from attrs import define, field

from moat.util import NotGiven, Path, PathLongener, PathShortener
from moat.util.exc import ExpKeyError

from ..meta import MsgMeta

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable, Iterator
    from typing import Any
    from moat.lib.cmd import Key

logger = getLogger(__name__)


def _keys_repr(x):
    return ",".join(str(k) for k in x)


@define
class Node:
    """Represents one MoaT-Link item."""

    _data: Any = field(init=False, default=NotGiven)
    _meta: MsgMeta | None = field(init=False, default=None)

    _sub: dict[Key,Node] = field(init=False, factory=dict, repr=_keys_repr)  # sub-entries

    def __attrs_post_init__(self, data: Any = NotGiven, meta: MsgMeta | None = None):
        if data is not NotGiven:
            self._data = data
            self._meta = meta

    def set(self, item: Path, data: Any, meta: MsgMeta, force: bool = False) -> bool | None:
        """Save new data below this node.

        If @tick is earlier than the item's timestamp, always return False.
        If data changes, apply change and return True.
        If @force is not set, return False.
        Otherwise, update metadata and return None.
        """
        assert isinstance(meta, MsgMeta)
        s = self.get(item)
        if s._meta is not None:  # noqa:SLF001
            if meta.timestamp <= s._meta.timestamp:  # noqa:SLF001
                return False
            if not force and s._data == data:  # noqa:SLF001
                return None
        s.set_(item,data,meta)
        return True

    def set_(self, path:Path, data:Any, meta:MsgMeta):
        "Low-level node data setter. The (sub)path is not stored by default."
        self._data = data
        self._meta = meta

    @property
    def data(self) -> Any:
        "return current data, raises ValueError if empty"
        if self._data is NotGiven:
            raise ValueError("empty node")
        return self._data

    @property
    def data_(self) -> Any:
        "return current data, returns NotGiven if empty"
        return self._data

    def keys(self):
        return self._sub.keys()

    def items(self):
        return self._sub.items()

    def __bool__(self) -> bool:
        "check if data exist"
        return self._data is not NotGiven

    def __eq__(self, other):
        if self._data != other._data:
            return False
        if self._sub != other._sub:
            return False
        return True

    def _dump_x(self):
        # Iterator that returns a serialization of this node tree.
        ps = PathShortener()
        for p, d, m in self._dump_x_(()):
            s, p = ps.short(p)
            yield (s, p, d, *(m.dump() if m is not None else ()))

    def _dump_x_(self, path):
        # Helper for _dump_x
        if self._data is not NotGiven:
            yield path, self._data, self._meta
        for k, v in self._sub.items():
            yield from v._dump_x_(
                path + (k,),
            )

    def dump(self):
        """
        An iterator that returns a path-shortened serialization of this
        node tree.
        """
        # The naïve method (in `_dump_x`) creates a full-path tuple for
        # each node, all of which the PathShortener will throw away.
        #
        # This code yields the exact same data – without that overhead.
        # The old code is kept (a) because it's more easily understood,
        # (b) for unit testing.

        yield from self._dump((), 0)

    def _dump(self, path, level):
        if self._data is not NotGiven:
            yield (level, path, self._data, *(self._meta.dump() if self._meta is not None else ()))
            level += len(path)
            path = ()
        for k, v in self._sub.items():
            it = iter(v._dump(path + (k,), level))
            try:
                d = next(it)
            except StopIteration:
                pass
            else:
                yield d
                level += len(path)
                path = ()
                yield from it

    def load(self, force=False):
        """
        receives a data stream created by `dump`.

        if @force is set, overwrite existing data even if newer.
        """
        # TODO mirror dump() and do this without a PathLongener
        pl = PathLongener()
        while True:
            s, p, d, *m = yield
            m = MsgMeta.restore(m, NotGiven)
            p = pl.long(s, p)
            n = self.get(p)
            if force or n.meta is None or n.meta.timestamp < m.timestamp:
                n._data = d
                n._meta = m

    @property
    def meta(self) -> MsgMeta|None:
        "return current metadata"
        return self._meta

    @meta.deleter
    def meta(self) -> None:
        "Clear metadata"
        del self._meta

    def __delitem__(self, item) -> None:
        """
        Remove an item. (It must be empty.)

        **Warning** Don't call this unless the timeout for deletion has passed.
        """
        d = self._sub[item]

        if d._sub or d.data is not NotGiven:
            raise ValueError(item)
        del self._sub[item]

    def __getitem__(self, item) -> Node:
        """Look up the entry.

        Raises KeyError if it doesn't exist.
        """
        if isinstance(item, Path):
            s = self
            for k in item:
                try:
                    s = s._sub[k]  # noqa:SLF001
                except KeyError:
                    raise ExpKeyError(k)
        else:
            s = self._sub[item]

        if s._data is NotGiven:  # noqa:SLF001
            raise ExpKeyError(item)
        return s

    def get(self, item, create=None) -> Node:
        """Look up an entry. Create if it doesn't exist.

        Unlike data[key], an "empty" key is not an error.
        """
        if item is Ellipsis:
            return self

        if isinstance(item, Path):
            s = self
            for n, k in enumerate(item):
                if isinstance(k, Path):
                    # import traceback
                    # logger.warning("Looking up %r\n%s", item, ''.join(traceback.format_stack()))
                    logger.warning("Looking up %r\n", item)
                    s = s.get(k, create=create)
                    continue
                try:
                    s = s._sub[k]  # noqa:SLF001
                except KeyError:
                    if create is False:
                        raise
                    s = s._add(k)  # noqa:SLF001
                else:
                    if create is True and n == len(item) - 1 and s._data is not NotGiven:
                        raise KeyError(k)
            return s

        try:
            res = self._sub[item]
        except KeyError:
            if create is False:
                raise
            return self._add(item)
        else:
            if create is True and res._data is not NotGiven:
                raise KeyError(item)
            return res

    def _add(self, item):
        if item in self._sub:
            raise ValueError("exists")
        self._sub[item] = s = type(self)()
        return s

    def __iter__(self) -> Iterator[tuple[Key, Node]]:
        """
        Return a list of keys under this node.
        """
        return iter(self._sub.items())

    def __contains__(self, item) -> bool:
        if isinstance(item, Path):
            s = self
            for k in item:
                try:
                    s = s._sub[k]  # noqa:SLF001
                except KeyError:
                    return False
            return True

        return item in self._sub

    def deleted(self) -> bool:
        """
        Check whether this tick has been marked as deleted.
        """
        return self._data is NotGiven

    async def walk(
        self,
        proc: Callable[[Path, Node], Awaitable[bool|None]],
        max_depth: int = 999999,
        min_depth: int = 0,
        timestamp: int | float = 0,
        depth_first: bool = False,
        force: bool = False,
    ):
        """
        Calls coroutine ``proc(node,Subpath)`` on this node and all its children.

        Deleted nodes are passed if they still have a Meta entry.

        if @depth_first is not set and @proc explicitly returns False,
        the subtree is skipped.

        if @force is set, also visit empty nodes.
        """

        async def _walk(s, p):
            if depth_first and max_depth > len(p):
                for k, v in s._sub.items():
                    await _walk(v, p / k)

            if min_depth <= len(p) and (force or (s.meta is not None and s.meta.timestamp >= timestamp)):
                if await proc(p, s) is False:
                    return

            if not depth_first and max_depth > len(p):
                for k, v in list(s._sub.items()):
                    await _walk(v, p / k)

        await _walk(self, Path())


    def search(self, path:Path) -> Node:
        """
        Find the destination node of a path, including wildcards.
        """
        nf = NodeFinder(self)
        for elem in path:
            nf.step(elem)
        return nf.result


class NodeFinder:
    """A generic object that can walk down a possibly-wildcard-equipped path.

    Example: given a path `one.two.three` and a root with subtree `*.three`,
    `NodeFinder(root).step(one).step(two).step(three)` will return the node
    at `*.three` (assuming that nothing more specific hangs off `the root`).

    If nothing is found, raises `KeyError`.
    """

    def __init__(self, src):
        self.steps = ((src, False),)

    def step(self, name, new=False):
        steps = []
        for node, keep in self.steps:
            if name in node:
                steps.append((node.get(name), False))
        for node, keep in self.steps:
            if "+" in node:
                steps.append((node.get("+"), False))
        for node, keep in self.steps:
            if "#" in node:
                steps.append((node.get("#"), True))
        for node, keep in self.steps:
            if keep:
                steps.append((node, True))
            # Nodes found with '*' stay on the list
            # so that they can match multiple path elements.
        if not steps:
            raise KeyError(name)
        self.steps = steps

    @property
    def result(self) -> tuple[Path,Node]:
        s = self.steps[0]
        return s[0]




"""
This module describes the MoaT data link.
"""

from __future__ import annotations

import os
import anyio
from platform import uname

# pylint: disable=missing-module-docstring
import logging
import sys
from functools import partial
from pathlib import Path as FSPath
import time

import asyncclick as click

from mqttproto import MQTTException

from moat.util import NotGiven, P, Path, load_subgroup, yprint, attrdict
from moat.util.path import set_root

from .backend import RawMessage, get_backend
from .client import Link

try:
    from mqttproto import MQTTPublishPacket
except ImportError:
    MQTTPublishPacket = NotGiven


logger = logging.getLogger(__name__)

usage1 = """
"moat link" requires configuration. It should look like this:

    link:
        backend:
            driver: mqtt
            host: localhost
            port: 1883
            user: foo
            pass: bar
            codec: std-cbor
        root: !P moat.something-unique

The MQTT hierarchy below the "root" topic, with slashes instead of dots,
*must* be empty when you start a new installation.

Please add this stanza to the file "/etc/moat/moat.cfg" (or "/etc/moat.cfg",
or "~/.config/moat.cfg") and try again.
"""

usage2 = """
"moat link" requires a root topic.

This entry should be tagged with '!P' and be a dot-separated path.
For clarity it should start with 'moat.', though that's just a
recommendation.

Your MQTT setup may NOT use any non-MoaT topics under that name
(replacing the dots with slashes).

Config example:

    link:
        backend:
            …
        prefix: !P moat.com.your-domain

"""

usage9 = """

"moat link" requires at least one history server for stable operation.

Please run "sudo systemctl start moat-link-server", or
start "moat link server" in a separate terminal, and try again.
"""


@load_subgroup(sub_pre="moat.link.cmd", sub_post="cli", ext_pre="moat.link", ext_post="_main.cli")
@click.option("-n","--name", type=str, help="Name of this client (or server)")
@click.pass_context
async def cli(ctx,name):
    """
    MoaT's data link

    This collection of commands is useful for managing and building MoaT itself.
    """
    obj = ctx.obj
    cfg = obj.cfg
    obj.name = name

    if "link" not in cfg or "backend" not in cfg["link"]:
        sys.stderr.write(usage1)
        raise click.UsageError("not configured")
    if not isinstance(cfg.link.root, Path) or cfg.link.root == P("XXX.NotConfigured.YZ"):
        sys.stderr.write(usage2)
        raise click.UsageError("badly configured")


@cli.command("host")
@click.option("-m","--main", is_flag=True, help="Main server flag (override)")
@click.option("-d","--debug", is_flag=True, help="Debug?")
@click.pass_obj
async def host(obj, main, debug):
    """Host management.

    This subcommand should run on each MoaT-Link connected host.

    It provide ping messages and related services.
    """
    from .host import cmd_host

    cfg = obj.cfg.link
    if obj.name is not None:
        raise click.UsageError("'moat link host' uses the hostname.")
    name = uname().node
    if not main:
        main = name == cfg.main
    cfg.backend.will = attrdict(
            topic=P("ping")/name,
            payload=dict(state="disconnected",up=False)
            )
    async with Link(cfg, name="!"+name) as link:
        await cmd_host(link, cfg, main=main, debug=debug)


@cli.command("test")
@click.pass_obj
async def test(obj):
    "Test"

    lock = anyio.Lock()
    cfg = obj.cfg.link
    set_root(cfg)
    async def check_root():
        try:
            with anyio.fail_after(1) as sc:
                async with back.monitor(cfg.root) as mon:
                    async for msg in mon:
                        async with lock:
                            print("# Retained root dataset:")
                            yprint(msg.data)
                            print("---")
                            return
        except TimeoutError:
            print(f"## No retained root dataset on {cfg.root}.")

    async def check_server():
        try:
            with anyio.fail_after(1) as sc:
                async with back.monitor(P(":R.run.service.main.conn")) as mon:
                    async for msg in mon:
                        async with lock:
                            print("# Server link:")
                            yprint(msg.data)
                            print("---")
                            return
        except TimeoutError:
            print(f"### No server link on {cfg.root}!")


    async with get_backend(cfg) as back,back.connect(), anyio.create_task_group() as tg:
        tg.start_soon(check_root)
        tg.start_soon(check_server)
        




"""
Placeholder for data that does schema verification
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.util import Path

__all__ = ["Data", "Schema"]


class Data(dict):
    "Schema-verified data (dict)"


class _SchemaName(str):
    __slots__ = ()

    def __getattr__(self, x):
        if len(self):
            return _SchemaName(f"{self}.{x}")
        return _SchemaName(x)


Schema = _SchemaName("")



"""
Gateway to wherever
"""

from __future__ import annotations

import anyio
from attrs import define,field
import logging

from moat.link.node import Node
from moat.link.meta import MsgMeta
from moat.lib.codec import get_codec
from moat.util import NotGiven, P, Path, to_attrdict

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .client import Watcher
    from moat.lib.codec import Codec

__all__ = ["Gate"]


class GateVanished(RuntimeError):
    "internal error: gate got dropped, or driver changed"
    pass

@define
class GateNode(Node):
    """
    A gatewayed node. It stores the external value and metadata.

    Data and meta 
    """
    ext_meta:dict[str,Any]|None=field(init=False, default=None)
    ext_data:Any=field(init=False, default=NotGiven)
    lock:anyio.abc.Lock=field(init=False, factory=anyio.Lock)

    todo:bool=field(init=False,default=False)

    @property
    def has_src(self):
        "Check whether source data is present"
        return self.data_ is not NotGiven or self.meta not in (None,NotGiven)

    @property
    def has_dst(self):
        "Check whether destination data is present"
        return self.ext_data is not NotGiven or self.ext_meta not in (None,NotGiven)

    @property
    def has_both(self):
        "Check whether both source and destination data are present"
        if self.data_ is NotGiven and self.meta is None:
            return False
        if self.ext_data is NotGiven and self.ext_meta is None:
            return False
        return True


class Gate:
    """
    This is the base class for data gateways.

    Gateways are described by a dict in ``:r.gate.NAME`` with the following
    entries:

    * src: source path, covered by ``moat.link``
    * dst: destination, *must not* be at or under the ``moat.link`` root (if MQTT)
    * driver:
      * mqtt: the destination is a raw MQTT thing
    * codec: Encoding of the destination (source is always ``std-cbor``).
    * retain: ``True/False/None``; the latter is the default and copies
      the data's retain flag

    The gateway works thus:
    * if a data item is not in the source or arrives from dest, copy to source
    * if a data item is not in the destination or arrives from source, copy to dest
    * if the values are equal, do nothing
    * if the source metadata say the data is from the destination, copy dest to source
    * otherwise copy source to dest.

    Subclasses override
    * get_dst
    * set_dst
    * newer_dst
    """

    state:Node
    src:Node
    tg:anyio.abc.TaskGroup
    codec:Codec

    _src_done:anyio.Event
    _dst_done:anyio.Event

    cfg:attrdict
    cf:attrdict

    def __init__(self, cfg:dict[str,Any], cf:dict[str,Any], path:Path, link:Link):
        """
        Setup.
        @cfg: initial data for this gateway.
        """
        self.cfg = cfg
        self.cf = to_attrdict(cf)
        self.codec = cf.get("codec", "cbor")
        if isinstance(self.codec, str):
            self.codec = get_codec(self.codec)

        self.link = link
        self.path = path
        self.origin = str(Path("GATE")+(P(cf["name"]) if "name" in cf else self.path[1:]))

        self.logger = logging.getLogger(f"moat.link.{path}")

    
    async def get_src(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        """
        Fetch the internal data.
        """
        async with self.link.d_watch(self.cf.src, subtree=True,state=None,meta=True,mark=True) as mon:
            task_status.started()
            async for pdm in mon:
                if pdm is None:
                    self._src_done.set()
                    continue
                p,d,m = pdm
                if m is not None and m.origin == self.origin and self._src_done.is_set():
                    # mine, so skip
                    continue

                node = self.data.get(p)
                if self.running or node.has_src:
                    # self.logger.debug("S NOW %r %r %r",p,d,m)
                    await self._set_dst(p,node,d,m)
                else:
                    # self.logger.debug("S DLY %r %r %r",p,d,m)
                    node.set_(p,d,m)
                    node.todo=True

    async def _set_dst(self, path:Path,node:GateNode,data:Any,meta:MsgMeta):
        node.ext_data=NotGiven
        node.ext_meta=NotGiven
        node.set_(path,data,meta)
        node.todo=False

        async with node.lock:
            await self.set_dst(path,data,meta,node)

    def dst_is_current(self):
        self._dst_done.set()

    async def get_dst(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        """
        Fetch the external data.

        Override this; call `set_src` with each item.

        You must call `dst_is_current` when the current state has been read
        and you're now waiting for updates. If your backend doesn't support
        this, use a timeout *and* an update counter; `set_src` returns True

        """
        raise NotImplementedError


    async def set_src(self, path:Path, data:Any, aux:MsgMeta):
        """
        Update source state (possibly). @aux is additional metadata that
        the destination resolver can use to disambiguate.
        """
        node = self.data.get(path)

        if self.running or node.has_dst:
            await self._set_src(self.cf.src+path,node,data,aux)
        else:
            node.ext_data = data
            node.ext_meta = aux or NotGiven
            node.todo = True

    async def _set_src(self, path:Path,node:GateNode,data:Any,aux:MsgMeta):
        async with node.lock:
            if not self.is_update(node,data,aux):
                return
            meta = MsgMeta(origin=self.origin)
            if aux not in (None,NotGiven):
                meta["gw"] = aux

            await self.link.d_set(path,data,meta)

            node.set_((),NotGiven,NotGiven)
            node.ext_data = data
            node.ext_meta = aux or NotGiven
            node.todo = False

    async def set_dst(self, path:Path, data:Any, meta:MsgMeta, node:GateNode):
        """
        Update destination state. @meta is the source metadata, in case
        it is useful in some way.
        """
        raise NotImplementedError

    def is_update(self, node:GateNode, data:Any,aux:MsgMeta):
        """
        Check whether this new destination data is an update.
        """
        return True

    def newer_dst(self, node) -> bool|None:
        """
        Test whether the destination data is newer, based on the node's
        metadata. Return `True` if the data should be copied to the source,
        `False` if the source should be copied to the destination, or
        `None` if inconclusive.

        This method is only called when starting up.
        """
        raise NotImplementedError

    async def run(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        """
        Run a bidirectional copy.

        This method auto-restarts the gateway if its data changes.
        It ends if the gateway node is removed or the driver changes.

        The task status is set when the initial sync has completed.

        Called by the system.
        """
        run = True
        while run:
            self.state = Node()
            self.data = GateNode()
            self._src_done=anyio.Event()
            self._dst_done=anyio.Event()
            self.running = False

            try:
                async with anyio.create_task_group() as self.tg:
                    await self.tg.start(self._restart)
                    await self.run_(task_status=task_status)
            except* GateVanished:
                run = False
            else:
                task_status=anyio.TASK_STATUS_IGNORED
                await anyio.sleep(1)


    async def _restart(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        "Restart the thing when the root changes."
        async with self.link.d_watch(self.path) as mon:
            task_status.started()
            async for d in mon:
                if self.cf == d:
                    continue
                if d is NotGiven or d.get("driver") != self.cf.driver:
                    raise GateVanished(str(self.path))
                self.cf = d
                self.tg.cancel_scope.cancel()
                return


    async def run_(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        """
        The core runner for the gateway.

        If your implementation needs a context or a support task,
        override this and call the original. `tg` can be used.
        """
        # start initial loops
        await self.tg.start(self.get_src)
        await self.tg.start(self.get_dst)

        # wait for initial scans to be done
        await self._src_done.wait()
        await self._dst_done.wait()
        self.running = True

        # resolve any conflicts in the initial data
        async def visit(path,node):
            if not node.todo:
                return

            if not node.has_src:
                # no source data
                if not node.has_dst:
                    # no destination data
                    return

                # copy dest to source
                d=True

            elif not node.has_dst:
                # copy source to dest
                d=False

            else:
                # both are set. Ugh.
                d = self.newer_dst(node)

            if d is False:
                self.logger.debug("SRC %s %s %r/%r",self.path,path, node.data_,node.meta)
                await self._set_dst(path,node,node.data_,node.meta)

            elif d is True:
                self.logger.debug("DST %s %s %r/%r",self.path,path, node.ext_data,node.ext_meta)

                meta = MsgMeta(origin=self.origin)
                if node.ext_meta:
                    meta["gw"] = node.ext_meta
                await self.link.d_set(self.cf.src+path,node.ext_data,meta)

            elif node.data_ != node.ext_data:
                self.logger.warning("Conflict %s %s %r/%r vs %r/%r",self.path,path, node.data_,node.meta,node.ext_data,node.ext_meta)

        await self.data.walk(visit, force=True)
        task_status.started()


    async def state_updater(self, mon:Watcher, *, task_status=anyio.TASK_STATUS_IGNORED):
        """
        Status update handler.

        By default this just gets the monitor node and updates the raw
        node data in the background.
        """
        node = await mon.get_node()
        task_status.started(node)

        # nothing further to do


async def run_gate(cfg: dict, link:Link, cf:Path|str, *, task_status=anyio.TASK_STATUS_IGNORED):
    """
    Run a gate in @link, described by @name.
    """
    from importlib import import_module

    if isinstance(cf,str):
        cf = P("gate")/cf
    path = cf
    cf = await link.d_get(path)

    drv = cf["driver"]
    if "." not in drv:
        drv = "moat.link.gate." + drv
    gate = import_module(drv).Gate(cfg, cf, path, link)
    await gate.run(task_status=task_status)



from __future__ import annotations

import anyio
import time
from attrs import define, field
from contextlib import nullcontext, AsyncExitStack
from moat.link import protocol_version
from collections import deque
from moat.util import as_service, P, srepr, attrdict, NotGiven
from moat.util.times import humandelta
import logging

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import NoReturn
    from moat.link.client import Link

__all__ = ["ntfy_bridge"]

logger = logging.getLogger(__name__)

@define
class IDT:
    """
    This object holds a single link ID.
    """
    id = field(type=str)
    # ID

    up = field(type=bool|None,kw_only=True,default=None)
    # None: no link state seen, True: latest link state

    host = field(type=str,default=None,kw_only=True)
    # host name seen

    id_seen = field(type=bool,default=False,init=False)
    # flag whether ID message seen

    last = field(factory=time.time, init=False)
    # time of last Ping (or other message)
    
    def __str__(self):
        return srepr(dict(id=self.id,up=self.up,ids=self.id_seen,host=self.host,last=humandelta(time.time()-self.last)))


class Mon:
    """
    Monitor IDs (and the host entries that point to them).
    """
    debug=False
    def __init__(self, cfg:dict, link:Link):
        self.cfg = cfg
        self.link = link

        self._ids = {}
        self._idq = deque()

    async def run(self, debug=False, *, task_status=anyio.TASK_STATUS_IGNORED):
        self.debug=debug
        async with anyio.create_task_group() as tg:
            await tg.start(self._mon_ping)
            await tg.start(self._mon_id)
            await tg.start(self._mon_host)
            await tg.start(self._timer)
            task_status.started()

            while debug:
                await anyio.sleep(5)
                print("IDs:")
                for idt in self._ids.values():
                    print(idt)

    async def _timer(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        # monitor _idq
        timeout = self.cfg.timeout.no_ping
        ping = self.cfg.timeout.ping
        task_status.started()
        while True:
            if not self._idq:
                # This should never happen except on startup, so ignore.
                await anyio.sleep(ping)
                continue

            while self._idq:
                t=time.time()
                tt,idt = self._idq.popleft()

                if tt+timeout > t:
                    await anyio.sleep(tt+timeout-t)
                if idt.up is False:
                    # already down
                    continue
                if tt+1<idt.last:
                    # new packet arrived
                    continue
                await self._cleanup(idt)

    async def _cleanup(self, idt:IDT):
        if self.debug:
            print("Clean",str(idt))

        if idt.up is False:
            return
        idt.up = False
        if idt.host is not None:
            await self.link.d_set(P("host")/idt.host, retain=True)
        if idt.id_seen:
            await self.link.d_set(P("run.id")/idt.id, retain=True)
        self._ids.pop(idt.id,None)

    async def _mon_ping(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        # Monitor ping messages
        path=P("run.ping.id")
        async with self.link.d_watch(path, subtree=True, state=False) as mon:
            task_status.started()
            async for p,msg in mon:
                if len(p) != 1:
                    logger.warning("Mon %s: ?? %s", path+p, msg)
                    continue

                id=p[0]
                if self.debug:
                    print("PING",id,srepr(msg))

                if (idt := self._ids.get(id, None)) is not None:
                    if not msg.get("up"):
                        # going down
                        await self._cleanup(idt)
                        continue
                    # mark seen
                    idt.up = True
                    idt.last = time.time()
                else:
                    if not msg.get("up"):  # down anyway. Ignore.
                        continue
                    self._ids[id] = idt = IDT(id=id, up=True)

                self._idq.append((time.time(),idt))

    async def _mon_host(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        path=P("host")
        async with self.link.d_watch(path, subtree=True, state=None) as mon:
            task_status.started()
            async for p,msg in mon:
                if len(p) != 1:
                    logger.warning("Mon_host %s: ?? %s", path+p, msg)
                    continue
                if msg is NotGiven:
                    continue

                host=p[0]
                id=msg["id"]
                if self.debug:
                    print("HOST",host,srepr(msg))

                if (idt := self._ids.get(id, None)) is None:
                    self._ids[id] = idt = IDT(id=id, up=None)
                elif idt.up is False:
                    idt.up = None  # owch?
                idt.host=host
                self._idq.append((time.time(),idt))

    async def _mon_id(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        path=P("run.id")
        async with self.link.d_watch(path, subtree=True, state=None) as mon:
            task_status.started()
            async for p,msg in mon:
                if msg is NotGiven:
                    continue
                if len(p) != 1:
                    logger.warning("Mon_id %s: ?? %s", path+p, msg)
                    continue

                id=p[0]
                if self.debug:
                    print("ID",id,srepr(msg))

                if (idt := self._ids.get(id, None)) is None:
                    self._ids[id] = idt = IDT(id=id, up=None)
                elif idt.up is False:
                    idt.up = None  # owch?
                idt.id_seen=True
                self._idq.append((time.time(),idt))


async def cmd_host(link:Link, cfg:dict, main:bool=False, *, debug=False, task_status=anyio.TASK_STATUS_IGNORED) -> NoReturn:
    """
    Host specific runner.

    This command tells MoaT-Link that a particular host is up.
    """

    async with AsyncExitStack() as ex:
        if task_status is anyio.TASK_STATUS_IGNORED:
            srv = await ex.enter_async_context(as_service(attrdict(debug=debug)))
            tg = srv.tg
        else:
            srv = task_status
            tg = await ex.enter_async_context(anyio.create_task_group())

        await link.d_set(P("host")/link.name, dict(id=link.id), retain=True)
        if main:
            await srv.tg.start(Mon(cfg, link).run, debug)
        srv.started()

        await anyio.sleep_forever()



from __future__ import annotations

import anyio
import httpx
import time
from contextlib import asynccontextmanager
from moat.link import protocol_version
from moat.util.misc import srepr
from moat.util import as_service, P, Path, attrdict

from typing import TYPE_CHECKING
from . import Notifier as BaseNotifier

if TYPE_CHECKING:
    from typing import NoReturn
    from moat.link.client import Link

__all__ = ["ntfy_bridge"]

prio_map = {
    "debug":1,
    "info":2,
    "warning":3,
    "error":4,
    "fatal":5,
}


class Notifier(BaseNotifier):
    @asynccontextmanager
    async def _ctx(self):
        async with (
            super()._ctx(),
            httpx.AsyncClient() as self.http,
        ):
            yield self

    async def send(self, topic:str|Path, title:str, msg:str, prio:str|None=None, tags:Sequence[str]=[], **kw):
        "Forward a message to NTFY"

        if "topic" in self.cfg:
            top = self.cfg.topic
        else:
            if isinstance(topic,str):
                topic=P(topic)
            try:
                top = topic[0]
            except IndexError:
                top = "TOP"
            else:
                topic = Path.build(topic[1:])

        url = self.cfg.url+"/"+top
        tags = list(tags)
        tags.append(str(topic))

        hdr = {}

        if title:
            hdr["title"] = title
        if prio is not None:
            try:
                prio = prio_map[prio]
                hdr["prio"] = str(prio)
                if prio > 3:
                    tags.append("warning")
            except KeyError:
                hdr["prio"] = "high"
                msg += f" (prio:{prio})"
        hdr["tags"] = ",".join(str(x) for x in tags)

        if "token" in self.cfg:
            hdr["authorization"] = f"Bearer {self.cfg['token']}"

        await self.http.post(url, data=msg, headers=hdr)



"""
MoaT gateway
"""

from __future__ import annotations

import anyio
from contextlib import AsyncExitStack

from moat.link.client import LinkCommon
from . import Gate as _Gate
from moat.util import Path, CFG, NotGiven, P
from moat.link.meta import MsgMeta

from typing import TYPE_CHECKING,overload                                                      
if TYPE_CHECKING:
    from moat.link.node.codec import CodecNode

class Gate(_Gate):
    codecs:CodecNode|None = None

    async def run_(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        "Main loop. Overridden to fetch the codecs"
        async with AsyncExitStack() as ex:
            if isinstance(self.cf.codec,Path):
                cdv = await ex.enter_async_context(self.link.d_watch(P("conv")+self.cf.codec, subtree=True,state=None,meta=False))
                self.codec_vecs = await cdv.get_node()

                self.codecs = await self.link.get_codec_tree()

            await super().run_(task_status=task_status)

    async def get_dst(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        async with AsyncExitStack() as ex:
            if self.codecs is not None:
                codec="noop"

                def conv(p,d):
                    # two step
                    # (a) look up the codec type in the vector
                    try:
                        vd = self.codec_vecs.search(p)
                        cd = self.codecs.get(Path.build(vd.data["codec"]))
                    except ValueError:
                        return NotGiven
                    try:
                        return cd.dec_value(d)
                    except Exception as exc:
                        self.logger.error("Decode: %s %r: %r", p,d, exc)
                        return NotGiven
                    
            else:
                codec=self.codec
                def conv(p,d):
                    return d

            mon = await ex.enter_async_context(self.link.monitor(self.cf.dst, subtree=True, codec=codec))
            task_status.started()
            ld = len(self.cf.dst)
            while True:
                try:
                    with anyio.fail_after(self.cf.get("timeout",0.5)):
                        msg = await anext(mon)
                except TimeoutError:
                    break
                p=Path.build(msg.topic[ld:])
                res = conv(p,msg.data)
                if res is NotGiven:
                    continue
                await self.set_src(p, res, msg.meta)
            self.dst_is_current()

            async for msg in mon:
                if msg.meta is not None and msg.meta.origin == self.origin:
                    # mine, so skip
                    continue
                p=Path.build(msg.topic[ld:])
                if msg.data == b'':
                    res = NotGiven
                else:
                    res = conv(p,msg.data)
                    if res is NotGiven:
                        continue
                await self.set_src(p, res, msg.meta)


    async def set_dst(self, path:Path, data:Any, meta:MsgMeta, node:GateNode):
        meta=MsgMeta(origin=self.origin,timestamp=meta.timestamp)
        if data is NotGiven:
            await self.link.send(self.cf.dst+path, b'', retain=True, codec="noop", meta=meta)
        elif self.codecs is not None:
            try:
                vd = self.codec_vecs.search(path)
                cd = self.codecs.get(Path.build(vd.data["codec"]))
            except (ValueError,KeyError) as exc:
                self.logger.error("No codec: %s %r",path,data)
                return
            res = cd.enc_value(data)
            if isinstance(res,(str,bytes,bytearray)):
                await self.link.send(self.cf.dst+path, res, retain=True, codec="noop",meta=meta)
            else:
                self.logger.error("Bad codec: %s %r > %r",path,data,res)

        else:
            await self.link.send(self.cf.dst+path, data, retain=True, codec=self.codec, meta=meta)

        node.ext_meta=meta

    def is_update(self, node:GateNode, data:Any, aux:MsgMeta):
        "update test"
        # if the old metadata match the new, it's not an update.
        try:
            if node.ext_meta.origin == aux.origin and node.ext_meta.timestamp == aux.timestamp:
                return False
        except (AttributeError,KeyError):
            pass
        return True

    def newer_dst(self,node):
        # If the external message has no metadata, it can't be from us,
        # thus assume it's newer.
        if not node.ext_meta:
            return True

        # If the internal and external metadata match, the message is from
        # us, so nothing to do.
        if self.origin == node.ext_meta.origin:
            return None

        # If the internal message has a copy of the outside metadata, it's
        # either unmodified or older. Test the data to be sure.
        if "gw" in node.meta:
            if node.meta["gw"] == node.ext_meta:
                return None if node.data_ == node.ext_data else True
            else:
                return True

        # Otherwise, if the external message is ours, it's old.
        if node.ext_meta.origin == self.origin:
            return False

        # if the timestamps are too close, there might be a problem.
        if abs(node.ext_meta.timestamp-node.meta.timestamp) < .1:
            return None

        # Otherwise use the message with the newer timestamp.
        return node.ext_meta.timestamp > node.meta.timestamp








"""
Bare-bones connection to a MoaT server
"""

from __future__ import annotations


from attrs import define, field
from moat.util import P
import anyio
from moat.lib.cmd.base import MsgHandler, MsgSender
from . import protocol_version as proto_version, protocol_version_min as proto_version_min
from .common import CmdCommon
from .auth import AuthMethod
import logging

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any
    from collections.abc import Awaitable
    from moat.lib.cmd.base import MsgHandler
    from moat.lib.cmd.msg import Msg
    from moat.lib.cmd import Key

logger = logging.getLogger(__name__)


class NotAuthorized(RuntimeError):
    pass


def _to_dict(x: list[AuthMethod]) -> dict[str, AuthMethod]:
    return {a.name: a for a in x}


@define
class Hello(CmdCommon):
    """
    This object handles the initial handshake between two MoaT links.

    Usage:

    * The ``handle`` method that you supplied to your `MsgHandler` must
      forward all incoming commands to `Hello.handler` while
      ``auth_data`' is `None`.

    * Call `Hello.run`.

    Note that due to the async nature of the protocol, additional commands
    may arrive even before `Hello.run` has returned.

    Negotiated auth data are in ``.auth_data``.
    """

    me: str | None = field(default=None)
    them: str | None = field(default=None)

    auth_data: Any = field(init=False, default=None)

    auth_in: dict[str, AuthMethod] = field(kw_only=True, default={}, converter=_to_dict)
    auth_out: dict[str, AuthMethod] = field(kw_only=True, default={}, converter=_to_dict)

    me_server: bool = field(default=False)
    they_server: bool = field(init=False, default=False)

    _sync: anyio.Event = field(init=False, factory=anyio.Event)
    _done: anyio.Event = field(init=False, factory=anyio.Event)

    # min and max protocol versions we might accept
    protocol_min: int = field(kw_only=True, default=proto_version_min)
    protocol_max: int = field(kw_only=True, default=proto_version)

    # negotiated protocol version
    protocol_version: int = field(init=False, default=0)
    hello_seen: anyio.Event = field(init=False, factory=anyio.Event)
    hello_a: tuple[Any, ...] = field(init=False, default=())
    hello_kw: dict[str, Any] = field(init=False, default={})

    def __init__(self, *a, **kw):
        super().__init__()
        self.__attrs_init__(*a, **kw)  # pyright:ignore
        if self.me_server and self.me is None:
            raise ValueError("A server must have a name")

    async def handle(self, msg: Msg, rcmd: list[Key], *prefix:Key) -> None:
        """
        Dispatch an incoming "hello" message
        """
        if prefix:
            raise NotImplementedError
        if rcmd.pop() != "i":
            raise ValueError("No Hello/Auth")
        if len(rcmd) == 1 and rcmd[0] == "hello":
            res = await self.do_hello(msg)
            await msg.result(res)
            return
        if len(rcmd) != 2 or rcmd[1] != "auth":
            raise ValueError("No Hello/Auth")

        if self.auth_data is not None:
            # Some other method already succeeded
            await msg.result(False)
            return
        a = self.auth_in.get(rcmd[0], None)
        if a is None:
            await msg.result(False)
        else:
            await a.handle(self, msg)

    def authorized(self, data: Any) -> bool:
        """
        Called by an auth method to indicate that authorization worked.

        Returns True if this method was the first to succeed.
        """
        if self.auth_data is not None:
            return False
        self.auth_data = data
        return True

    doc_i_hello = dict(
        _d="Process remote Hello msg",
        _r="auth state",
        _0="int:protocol",
        _1="bool:server flag",
        _2="str:sender's name",
        _3="str:recipient's name (temp by sender)",
        _4=["str:auth method"],
        _k="str:auth names",
        _kw="Any:auth params",
    )

    async def do_hello(self, msg) -> bool | None:
        """
        Process the remote hello message.

        Returns True if no auth is required.
        """
        try:
            res = await self._do_hello(msg)
        except BaseException:
            self.auth_data = False
            raise
        else:
            if self.auth_data is None:
                self.auth_data = res
            await msg.result(res)
        finally:
            self._done.set()

    async def _do_hello(self, msg) -> bool | dict:
        logger.debug("H IN %r %r", msg.args, msg.kw)
        it = iter(msg.args)
        auth = True
        aux_data = {}

        try:
            prot = next(it)
            if prot < self.protocol_min:
                raise ValueError("Protocol mismatch", prot)
            self.protocol_version = min(prot, self.protocol_max)

            # TODO special auth for servers?
            self.they_server = next(it)
            if not self.they_server and not self.me_server:
                raise RuntimeError("Two clients cannot talk")

            # Remote names
            # If the remote is nameless, use ours.
            # If the remote is a server, keep its name.
            # If the name are the same, no problem.
            # If the remote name starts with an underscore, tell it to use ours.
            # If the remote name starts with a bang, strip it.
            # Otherwise prefix with our name.
            remote_name = next(it)
            if remote_name is None:
                if self.them is None:
                    logger.error("No remote name")
                    auth=False
                    raise StopIteration
                aux_data["name"] = self.them
            elif self.them is None:
                self.them = remote_name
            elif self.them != remote_name:
                if remote_name.startswith("_"):
                    logger.debug("Remote name: %r / %r", remote_name, self.them)
                    aux_data["name"] = self.them
                elif self.they_server:
                    self.them = remote_name
                elif remote_name[0] == "!":
                    self.them = remote_name[1:]
                else:  # we're the server, so use our name as a prefix
                    self.them = f"{self.me}.{remote_name}"
                    logger.debug("Remote name: %r", self.them)
                    aux_data["name"] = self.them

            local_name = next(it)
            if local_name is None:
                pass
            elif self.me is None:
                self.me = local_name
            elif self.me != local_name:
                logger.debug("My name: %r / %r", local_name, self.me)
                if not self.me_server:
                    self.me = local_name

            auth = next(it)

            if not next(it):
                raise RuntimeError("Not talking to a server")

        except StopIteration:
            pass

        # wait for the outgoing part to start
        await self._sync.wait()

        if auth is False:
            raise NotAuthorized("Remote blocks us", self.them)
        if auth is True:
            self.auth_data = True
            return aux_data or True

        if isinstance(auth, str):
            auth = (auth,)

        # Check for auth data in the Hello
        for a in self.auth_in.values():
            res = await a.hello_in(self, msg.kw.get(a.name, None))
            if res is False:
                return False
            if res:
                if self.auth_data is None:
                    self.auth_data = True
                break

        # cycle through the remote side's accepted auth methods
        for a in auth:
            am = self.auth_out.get(a, None)
            if am is None:
                continue
            res = await am.chat(self, self.hello_kw.get(a, None))
            if res is None:
                continue
            if isinstance(res,dict):
                aux_data.update(res)
            if res is False or not aux_data:
                return res
            return aux_data

        # Nothing matched.
        return False

    async def run(self, sender: MsgSender, **kw):
        """
        Send our Hello message.
        """

        auths = []
        for a in self.auth_in.values():
            auths.append(a.name)
            if a.name not in kw:
                v = await a.hello_out()
                if v is not None:
                    kw[a.name] = v

        if len(auths) == 0:
            auths = True
        elif len(auths) == 1:
            auths = auths[0]

        logger.debug("H OUT %d %s %s %r %r", proto_version, self.me, self.them, auths, kw)
        self._sync.set()
        res = await sender.cmd(
            P("i.hello"),
            proto_version,
            self.me_server,
            self.me,
            self.them,
            auths,
            **kw,
        )
        res = res.kw or res[0]

        if res is False:
            raise NotAuthorized("Server %r rejects us", self.them)

        # Wait for the incoming side of the auth/hello dance to succeed
        await self._done.wait()
        return res



# command line interface
from __future__ import annotations

import datetime
import time
import anyio

import asyncclick as click

from moat.util import MsgReader, NotGiven, P, PathLongener, attr_args, yprint, Path
from moat.util.times import ts2iso, humandelta

from moat.link.client import Link
from moat.link._data import data_get, node_attr
from moat.link.meta import MsgMeta
from moat.link.node import Node


@click.group(short_help="Manage notifications.")  # pylint: disable=undefined-variable
@click.pass_context
async def cli(ctx):
    """
    Handle notifications.
    """
    obj = ctx.obj
    cfg = obj.cfg["link"]
    obj.conn = await ctx.with_async_resource(Link(cfg, name=obj.name))



@cli.command()
@click.option("-b","--backend",type=str,multiple=True,help="Restrict to this backend")
@click.pass_obj
async def run(obj, backend):
    """
    Forward notification messages.

    This command monitors the 'notify' subpath and forwards messages to
    your ntfy.sh instance.
    """
    from moat.link.notify import Notify

    cfg = obj.cfg.link.notify
    if backend:
        cfg.backends = backend
    await Notify(cfg).run(obj.conn)




"""
Moat-Link is a moderately shallow library that distributes basic
topic-based messaging between several connections or encodings.

The point of this library is to support a minimal unified pub/sub messaging
service.

Configuration looks like this:

    link:
      dist:
        hass:
          prefix: !P hass
          backend: mqtt
          path: !P some.special.hass
          chop: 1
          codec: home_assistant

        root:
          prefix: !P dist.root
          backend: mqtt
          codec: cbor

        special:
          prefix: !P :
          backend: mqtt
          path: !P some.special.area
          chop: 0
          codec: cbor

        storage:
          prefix: !P :
          chop: 0
          backend: kv
          codec: null
          retained: true

      backend:
        mqtt:
          uri: mqtt://localhost:51883

"""

from __future__ import annotations

try:
    from importlib.metadata import version

    _version = version("moat.link")
    _version_tuple = tuple(int(x) for x in _version.split("."))

except Exception:  # pragma: no cover
    _version = "0.0.1"
    _version_tuple = (0, 0, 1)

protocol_version_min = 0
protocol_version = 0

__path__ = __import__("pkgutil").extend_path(__path__, __name__)



"""
Connection and command helpers
"""

from __future__ import annotations

from contextlib import asynccontextmanager
from moat.lib.cmd.anyio import run as run_stream
import anyio
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import AsyncContextManager
    from moat.lib.cmd import MsgSender
    from moat.lib.codec import Codec


@asynccontextmanager
async def TCPConn(cmd: MsgSender | None, *a, codec: str|Codec = "std-cbor", debug:bool=False, logger=None, **kw) -> AsyncGenerator[..., MsgSender]:
    """
    Connection to a MoaT server.

    This encapsulates a TCP link to a remote side.

    Parameters:
    * the MsgHandler to connect to
    * all other arguments go to `anyio.connect_tcp`

    This is an async context manager. It forwards incoming commands to @cmd
    and yields a `MsgSender` that forwards local to the remote side.
    """
    async with (
        await anyio.connect_tcp(*a, **kw) as stream,
        run_stream(cmd, stream, codec=codec, debug=debug, logger=logger) as hdl,
    ):
        yield hdl

@asynccontextmanager
async def UnixConn(cmd: MsgSender | None, *a, codec: str|Codec = "std-cbor", debug:bool=False, logger=None, **kw) -> AsyncGenerator[..., MsgSender]:
    """
    Connection to a MoaT server.

    This encapsulates a Unix-domain client link.

    Parameters:
    * the MsgHandler to connect to
    * all other arguments go to `anyio.connect_unix`

    This is an async context manager. It forwards incoming commands to @cmd
    and yields a `MsgSender` that forwards local to the remote side.
    """
    async with (
        await anyio.connect_unix(*a, **kw) as stream,
        run_stream(cmd, stream, codec=codec, debug=debug, logger=logger) as hdl,
    ):
        yield hdl



from __future__ import annotations

from moat.link import protocol_version

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import ClassVar, ReadOnly, Any
    from moat.lib.cmd import MsgHandler, Msg, Key
    from .hello import Hello

__all__ = ["AuthMethod", "TokenAuth", "AnonAuth", "NoAuth"]


class AuthMethod:
    name: ClassVar[str]

    async def hello_out(self) -> Any|None:
        """
        Return data that should be included in the Hello message we
        send.

        The default is no data.
        """
        return None

    async def hello_in(self, conn: Hello, data: Any) -> bool | None:
        """
        This may implement "The remote side authorizes itself to me" based
        on data sent in the remote's Hello message.

        If this returns False, the auth call fails; if True, other auth
        methods are skipped.

        Uf the remote uses the current protocol level and there is data,
        this will fail. Otherwise, do nothing.
        """
        if data is None or conn.protocol_version < protocol_version:
            return None

        return False

    async def chat(self, conn: Hello, data: Any):
        """
        The recipient of a Hello message whose ``auth`` member includes our
        name calls this method. It's supposed to call ``conn.cmd(i.auth.NAME), …)``
        (or a streaming version thereof) and return its eventual result.

        This method implements "I authorize me to the remote side".

        The default is to do nothing.
        """
        return None

    async def handle(self, conn: Hello, msg: Msg) -> None:
        """
        The dispatcher calls this method with an incoming ``i.auth.NAME`` message.

        It thus implements "The remote side authorizes itself to me".

        The default is to fail, because the remote shouldn't call us without reason.
        """
        await msg.result(False)


class TokenAuth(AuthMethod):
    name: ClassVar[ReadOnly[str]] = "token"

    def __init__(self, *token: str):
        self._token = token

    async def hello_out(self):
        return self._token[0] if self._token else None

    async def hello_in(self, conn, data):
        """
        Check the incoming token.
        """
        # We don't actually chat here.
        # The remote should have sent the token in its Hello message.

        if data is None:
            # client didn't send data: try another method
            return None

        if data in self._token:
            # save the token that was actually used
            # returns False if some other auth method succeeded first
            return conn.authorized(data) or None

        # wrong token: kick them off
        return False

    async def handle(self, conn: MsgHandler, msg: Msg):
        """
        The client shouldn't send an `i.auth.token` message.
        """
        await msg.result(False)


class AnonAuth(AuthMethod):
    """
    Auth method of last resort: anonymous login.
    """

    name: ClassVar[ReadOnly[str]] = "anon"

    async def hello_out(self):
        return None

    async def chat(self, conn, data):
        conn.authorized(self)
        return True

    async def handle(self, conn: MsgHandler, msg: Msg):
        return True


class NoAuth(AuthMethod):
    """
    Reject auth attempts.
    """

    name: ClassVar[ReadOnly[str]] = "no"

    async def hello_out(self):
        return None

    async def chat(self, conn, data):
        "reject"
        return False

    async def handle(self, conn: MsgHandler, msg: Msg, *prefix:Key):
        "reject"
        return False



from __future__ import annotations

import anyio
import time
from abc import ABCMeta, abstractmethod
from contextlib import AsyncExitStack, asynccontextmanager
from moat.link import protocol_version
from moat.util.misc import srepr
from moat.util import as_service, P, Path, attrdict, CtxObj
import logging

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import NoReturn
    from moat.link.client import Link

__all__ = ["ntfy_bridge"]

logger = logging.getLogger(__name__)

class Notify:
    def __init__(self, cfg):
        self.cfg = cfg

    async def run(self, link:Link):
        self.link = link
        async with AsyncExitStack() as ex:
            backend = self.cfg.backend
            if isinstance(backend,str):
                backend=(backend,)

            self._backends = {}
            for name in backend:
                cfg = self.cfg.get(name,attrdict())
                try:
                    notifier = await ex.enter_async_context(get_backend(name, link,cfg))
                except Exception as exc:
                    logger.warning("Backend %r", name, exc_info=exc)
                else:
                    self._backends[name] = notifier

            if not self._backends:
                raise RuntimeError("No backend worked.")

            try:
                await self._run()
            finally:
                with anyio.move_on_after(2, shield=True):
                    await self.send("error.notify",f"Backend stopped", "The backend terminated.", prio="fatal")


    async def send(self, *a, **kw) -> None:
        """Send this notification to that topic."""
        bad = []
        dropped = []
        for name,b_e in list(self._backends.items()):
            try:
                await b_e.send(*a,**kw)
            except Exception as exc:
                logger.warning("Backend %r", name, exc_info=exc)
                bad.append(b_e)

        for name in bad:
            if self._backends.pop(name,None) is not None:
                dropped.append(name)
        if not self._backends:
            raise RuntimeError("All backends failed.")

        for name in dropped:
            await self.send("error.notify",f"Backend {name} error", "The backend errored and was removed.", prio="error")


    async def _run(self) -> NoReturn:
        """
        A bridge that monitors the Notify subchannel and forwards to NTFY.SH
        (or a private server).
        """
        async with as_service(attrdict(debug=False)) as srv:
            try:
                await srv.tg.start(self._keepalive)

                async with self.link.d_watch(self.cfg.path,subtree=True,meta=True,state=None) as mon:
                    srv.set()
                    async for path,msg,meta in mon:
                        t = time.time()
                        if meta.timestamp < t-self.cfg.max_age:
                            continue
                        if isinstance(msg,dict):
                            if "title" not in msg:
                                title = str(path)
                            await self.send(topic=path, **msg)
                        else:
                            await self.send(topic=path, msg=str(msg))

                    # not reached, loop doesn't terminate

            except Exception as exc:
                await self.send(topic=P("error.notify"),
                    msg=repr(exc), prio="error",title="Gateway failure")
                raise


    async def _keepalive(self, *, task_status=anyio.TASK_STATUS_IGNORED) -> NoReturn:
        "Monitor the keepalive topic"
        bad = False
        link = self.link
        main_host = link.cfg.main
        keep = self.cfg.keepalive
        ok_keep = keep.ok

        timeout = keep.get("timeout",link.cfg.timeout.ping*1.5)

        async with link.d_watch(P("host")/main_host, state=None) as mon:
            task_status.started()
            mon = aiter(mon)
            while True:
                if bad:
                    msg = await anext(mon)
                else:
                    try:
                        with anyio.fail_after(timeout*2):
                            msg = await anext(mon)
                    except TimeoutError:
                        msg = None

                if isinstance(msg,dict) and "id" in msg:
                    async with link.d_watch(P("run.ping.id")/msg["id"]) as mon2:
                        mon2 = aiter(mon2)
                        while True:
                            try:
                                with anyio.fail_after(timeout):
                                    msg = await anext(mon2)
                                    if not msg.get("up", False):
                                        break
                                    if bad and "msg" in ok_keep:
                                        await self.send(topic="error.notify",**ok_keep)
                                        bad = False
                            except TimeoutError:
                                break

                res = await self.send(topic="error.notify", **keep)
                bad=True



def get_backend(name:str, link:Link, cfg: dict) -> Backend:
    """
    Fetch the backend named in the config and initialize it.
    """
    from importlib import import_module

    if "." not in name:
        name = "moat.link.notify." + name
    return import_module(name).Notifier(link,cfg)

class Notifier(CtxObj, metaclass=ABCMeta):
    "Base class for notification backends"

    def __init__(self, link:Link, cfg: attrdict):
        self.cfg = cfg
        self.link = link

    @asynccontextmanager
    async def _ctx(self) -> AsyncIterator[Self]:
        yield self

    @abstractmethod
    async def send(self, topic:str|Path, title:str, msg:str, **kw):
        """Send this notification to that topic."""



"""
This module describes the MoaT data link.
"""

from __future__ import annotations

import anyio

# pylint: disable=missing-module-docstring
import logging
import sys
from functools import partial
from pathlib import Path as FSPath
import time

import asyncclick as click

from mqttproto import MQTTException

from moat.util import NotGiven, P, Path, load_subgroup, yprint, gen_ident, PathLongener
from moat.util.path import set_root
from moat.util.times import ts2iso,humandelta
from moat.lib.codec import get_codec

from moat.link.backend import RawMessage, get_backend

try:
    from mqttproto import MQTTPublishPacket
except ImportError:
    MQTTPublishPacket = NotGiven


logger = logging.getLogger(__name__)

@click.group
@click.pass_context
async def cli(ctx):
    """
    Commands that bypass the MoaT-Link server

    These access the MQTT backend directly.
    """
    obj = ctx.obj
    cfg = obj.cfg["link"]
    if not obj.name:
        obj.name = "mon_"+gen_ident()

    set_root(cfg)
    obj.conn = await ctx.with_async_resource(get_backend(cfg, name=obj.name))


@cli.command()
@click.argument("file", type=click.Path(), nargs=1)
async def init(file):
    """Write an initial preload file.

    Usage: moat link raw init <node> <outfile>

    Writes an initial MoaT-Link data file.

    Using this command, followed by "moat kv server -l <outfile> <node>", is
    equivalent to running "moat kv server -i 'Initial data' <node>.
    """
    from moat.link.meta import MsgMeta
    from moat.util.msg import MsgWriter
    from moat.util.cbor import gen_start,gen_stop

    meta=MsgMeta(origin="init")
    async with MsgWriter(path=file, codec="std-cbor") as f:
        await f(gen_start("Initial link data file",mode="init"))
        await f([0,P(":"),dict(type="MoaT-Link data"),*meta.dump()])
        await f(gen_stop(mode="log_end"))

@cli.command("test")
@click.pass_obj
async def test(obj):
    "Test"

    lock = anyio.Lock()
    cfg = obj.cfg.link
    async def check_root():
        try:
            with anyio.fail_after(1) as sc:
                async with back.monitor(cfg.root) as mon:
                    async for msg in mon:
                        async with lock:
                            print("Root dataset:")
                            yprint(msg.data)
                            print("---")
                            return
        except TimeoutError:
            print(f"No retained root dataset on {cfg.root} found.")

    async def check_server():
        try:
            with anyio.fail_after(1) as sc:
                async with back.monitor(cfg.root+P("run.service.main")) as mon:
                    async for msg in mon:
                        async with lock:
                            print("Server broadcast:")
                            yprint(msg.data)
                            print("---")
                            return
        except TimeoutError:
            print(f"No retained root dataset on {cfg.root} found.")


    async with get_backend(cfg) as back,back.connect(), anyio.create_task_group() as tg:
        tg.start_soon(check_root)
        tg.start_soon(check_server)
        


def _get_message(args):
    for m in args["msg"]:
        yield m
    for m in args["msg_eval"]:
        yield eval(m)  # pylint: disable=eval-used
    if args["msg_lines"]:
        with open(args["msg_lines"]) as f:  # pylint: disable=unspecified-encoding
            for line in f:
                yield line.encode(encoding="utf-8")
    if args["msg_stdin_lines"]:
        for line in sys.stdin:
            if line:
                yield line.encode(encoding="utf-8")
    if args["msg_stdin"]:
        yield sys.stdin.buffer.read()
    if args["msg_stdin_eval"]:
        message = sys.stdin.read()
        yield eval(message)  # pylint: disable=eval-used


async def do_pub(client, args, cfg, codec):
    logger.info("%s Connected to broker", client.name)
    for k, v in args.items():
        if v is None or v is NotGiven:
            continue
        cfg[k] = v

    try:
        topic = args["topic"]
        retain = args["retain"]
        qos = args["qos"] or cfg["qos"]

        async with anyio.create_task_group() as tg:
            for message in _get_message(args):
                logger.info("%s Publishing to '%s'", client.name, topic)
                tg.start_soon(partial(client.send, topic, message, qos=qos, retain=retain, codec=codec))
        logger.info("%s Disconnected from broker", client.name)
    except KeyboardInterrupt:
        logger.info("%s Disconnected from broker", client.name)


@cli.command()
@click.option("-i", "--client_id", "--name", "name", help="string to use as client ID")
@click.option("-q", "--qos", type=click.IntRange(0, 2), help="Quality of service to use (0-2)")
@click.option("-r", "--retain", is_flag=True, help="Set the Retain flag")
@click.option("-t", "--topic", type=P, required=True, help="Message path")
@click.option("-m", "--msg", multiple=True, help="Message data (may be repeated)")
@click.option(
    "-M",
    "--msg-eval",
    multiple=True,
    help="Message data (Python, evaluated, may be repeated)",
)
@click.option(
    "-f",
    "--msg-lines",
    type=click.File("r"),
    help="File with messages (each line sent separately",
)
@click.option("-R", "--msg-stdin", is_flag=True, help="Single message from stdin")
@click.option(
    "-s",
    "--msg-stdin-lines",
    is_flag=True,
    help="Messages from stdin (each line sent separately",
)
@click.option(
    "-S",
    "--msg-stdin-eval",
    is_flag=True,
    help="Python code that evaluates to the message on stdin",
)
@click.option("-k", "--keep-alive", type=float, help="Keep-alive timeout (seconds)")
@click.option("-c", "--codec", type=str, default="noop", help="codec to use (default: noop)")
@click.pass_obj
async def pub(obj, **args):
    """Publish one or more MoaT-Link messages.

    This command directly accesses the MQTT server: it works without a
    working MoaT-Link server.
    """
    if args["msg_stdin"] + args["msg_stdin_lines"] + args["msg_stdin_eval"] > 1:
        raise click.UsageError("You can only read from stdin once")
    cfg = obj.cfg.link
    name = args["name"] or cfg.get("name", None)
    codec = get_codec(args["codec"])

    if args["keep_alive"]:
        cfg["keep_alive"] = args["keep_alive"]

    if not obj.name:
        raise UsageError("You must supply a client name")

    async with anyio.create_task_group() as tg:
        await do_pub(obj.conn, args, cfg, codec)

async def run_kvsub(client, topic, lock):
    """Monitor a MoaT-KV subtree"""

    if topic[-1] == "#":
        topic = topic.parent
        depth=-1
    else:
        depth=0
    async with client.watch(
        topic,
        nchain=2,
        fetch=True,
        max_depth=depth,
        long_path=False,
    ) as res:
        pl = PathLongener(topic)
        async for r in res:
            async with lock:
                tm = time.time()
                atm = anyio.current_time()

                #if "value" in r:
                #    add_dates(r.value)
                pl(r)
                if r.get("state", "") == "uptodate":
                    continue
                del r["seq"]

                r["_prev"] = humandelta(atm-lock.tm,msec=6)
                r["_time"] = ts2iso(tm, msec=6)
                r["time"] = tm

                yprint(r)
                print("---")
                lock.tm=atm

async def do_sub(client, args, cfg):
    "handle subscriptions"
    lock = anyio.Lock()
    lock.tm = anyio.current_time()
    try:
        async with anyio.create_task_group() as tg:
            for topic in args["topic"]:
                tg.start_soon(run_sub, client, topic, args, cfg.link, lock)
            if args["kv_topic"]:
                from moat.kv.client import open_client as kv_client
                async with kv_client(**cfg.kv) as kvc:
                    for topic in args["kv_topic"]:
                        tg.start_soon(run_kvsub, kvc, topic, lock)



    except KeyboardInterrupt:
        pass
    except MQTTException as ce:
        logger.fatal("connection to '%s' failed: %r", args["uri"], ce)


async def run_sub(client, topic, args, cfg, lock):
    "handle a single subscription"
    qos = args["qos"] or cfg["qos"]
    max_count = args["n_msg"]
    count = 0
    dcbor=get_codec("std-cbor")
    dmsgpack=get_codec("std-msgpack")

    async with client.monitor(topic, qos=qos, codec=args.get("codec","noop")) as subscr:
        async for msg in subscr:
            async with lock:
                if args["yaml"]:
                    tm = time.time()
                    atm = anyio.current_time()

                    d = dict(topic = msg.topic, time=tm, _prev = humandelta(atm-lock.tm,msec=6),_time = ts2iso(tm, msec=6))

                    if isinstance(msg, RawMessage):
                        d["raw"] = msg.data
                        d["error"] = repr(msg.exc)
                    else:
                        d["data"] = msg.data
                    if msg.meta is not None:
                        d["meta"] = msg.meta.repr()

                    try:
                        d["data_cbor"] = dcbor.decode(msg.data)
                    except Exception:
                        pass

                    try:
                        d["data_msgpack"] = dmsgpack.decode(msg.data)
                    except Exception:
                        pass

                    flags = ""
                    if msg.retain:
                        flags += "R"
                    #if msg.qos > 0:
                    #    flags += f"Q{int(msg.qos)}"
                    if flags:
                        d["_flags"] = flags

                    lock.tm=atm
                    yprint(d)
                    print("---")
                else:
                    if isinstance(msg, RawMessage):
                        print(msg.topic, "*", msg.data, repr(msg.exc), sep="\t")
                    else:
                        print(msg.topic, msg.data, sep="\t")
            count += 1
            if max_count and count >= max_count:
                break


@cli.command()
@click.option("-i", "--client_id", "--name", "name", help="string to use as client ID")
@click.option("-q", "--qos", type=click.IntRange(0, 2), help="Quality of service to use (0-2)")
@click.option("-y", "--yaml", is_flag=True, help="Print output as YAML stream")
@click.option(
    "-t",
    "--topic",
    multiple=True,
    type=P,
    help="Message topic, dot-separated (can be used more than once)",
)
@click.option(
    "-T",
    "--kv-topic",
    multiple=True,
    type=P,
    help="Message topic from MoaT-KV legacy, dot-separated (can be used more than once)",
)
@click.option("-n", "--n_msg", type=int, default=0, help="Number of messages to read (per topic)")
@click.option("-k", "--keep-alive", type=float, help="Keep-alive timeout (seconds)")
@click.pass_obj
async def sub(obj, **args):
    """
    Monitor one or more MoaT-Link topics.

    This command directly accesses the MQTT server: it works without a
    working MoaT-Link server.
    """
    cfg = obj.cfg

    name = args["name"] or cfg.link.get("id", None)

    if args["keep_alive"]:
        cfg.link["keep_alive"] = args["keep_alive"]

    await do_sub(obj.conn, args, cfg)



# command line interface
from __future__ import annotations

import datetime
import time
import anyio

import asyncclick as click
from moat.util import MsgReader, NotGiven, P, PathLongener, attr_args, yprint, Path, process_args
from moat.util.times import ts2iso, humandelta

from moat.link.client import Link
from moat.link._data import data_get, node_attr
from moat.link.meta import MsgMeta
from moat.link.node import Node


@click.command(short_help="Send a command")
@click.option("-S","--stream", is_flag=True, help="Read a stream")
@click.option("-R","--raw", is_flag=True, help="Show raw message data")
@click.argument("path", type=P, nargs=1)
@click.pass_context
@attr_args
async def cli(ctx, path, stream, raw, **kw):
    """
    Send a command to the server.
    """
    obj = ctx.obj
    cfg = obj.cfg["link"]
    if obj.port is not None:
        cfg.client.port = obj.port

    async with Link(cfg, name=obj.name) as conn:
        val = process_args(NotGiven, **kw)
        kw = {}
        if isinstance(val,list):
            args = val
        elif isinstance(val,dict):
            args = []
            for k,v in list(val.items()):
                if k.startswith("_"):
                    try:
                        kk = int(k[1:])
                    except ValueError:
                        pass
                    else:
                        if k>=len(args):
                            args += [NotGiven]*(k-len(args)+1)
                        args[k]=v
                        continue
                kw[k] = v
        elif val is NotGiven:
            args = []
        else:
            args = [val]

        if stream:
            async with conn.cmd(path, *args, **kw).stream_in() as res:
                yprint(rep_(res,raw), stream=obj.stdout)
                print("---", file=obj.stdout)
                async for msg in res:
                    yprint(rep_(msg,raw), stream=obj.stdout)
                    print("---", file=obj.stdout)
        else:
            res = await conn.cmd(path, *args, **kw)
        yprint(rep_(res,raw), stream=obj.stdout)


def rep_(res,raw=False):
    if raw:
        return res.args+[res.kw]
    if not res.args:
        return res.kw
    if res.kw:
        return res.args+[res.kw]
    if len(res.args) == 1:
        return res.args[0]
    return res.args



"""
This module contains the MoaT-Link nodes used as codecs.
"""

from __future__ import annotations

from logging import getLogger

from attrs import define, field

from moat.util import Path, make_proc

from . import Node

logger = getLogger(__name__)

@define
class CodecNode(Node):
    """
    This node type is used as a codec.
    """
    _enc = field(init=False,default=None)
    _dec = field(init=False,default=None)

    def enc_value(self, value, entry=None, **kv):
        if self._enc is not None:
            value = self._enc(value=value, entry=entry, data=self._data, **kv)
        return value

    def dec_value(self, value, entry=None, **kv):
        if self._dec is not None:
            value = self._dec(value=value, entry=entry, data=self._data, **kv)
        return value

    def set_(self, path:Path, data:Any, meta:MsgMeta):
        super().set_(path,data,meta)

        enc = None
        dec = None
        if data is not None and data.get("decode",None) is not None:
            dec = make_proc(data["decode"], ("value",), path)

        if data is not None and data.get("encode",None) is not None:
            enc = make_proc(data["encode"], ("value",), path)

        self._enc = enc
        self._dec = dec



"""
Encoding and decoding the metadata for MoaT
"""

from __future__ import annotations

import time
from base64 import b85decode, b85encode

from attrs import define, field
import ruyaml as yaml

from moat.lib.codec.proxy import as_proxy
from moat.util import NotGiven
from moat.util.cbor import StdCBOR
from moat.util.times import ts2iso

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Self, Literal
    from types import EllipsisType

_codec = StdCBOR()

SafeRepresenter = yaml.representer.SafeRepresenter  # pyright:ignore


def _gen(i):
    ii = i + 1

    def get_(self):
        self._len(ii)
        return self[i]

    def set_(self, val):
        self._len(ii)
        self[i] = val

    return property(get_, set_)


@as_proxy("_MM")
@define(kw_only=True)
class MsgMeta:
    """
    This class encapsulates a message's metadata by encoding them in a
    hopefully-space-saving way.

    XXX: either create a dictionary for origin names, as there aren't many,
    or rely on data compression?

    Currently defined offsets:

    * 0: origin: a string declaring which subsystem created a message.
    * 1: timestamp: the Unix timestamp when the message was originally
         created.

    You can use indexing to address any other array or keyword value.
    Deleting an array member sets it to ``None``.

    The "source" attribute is used for preventing loops within a program.
    It is not saved.
    """

    a: list[Any] = field(factory=list)
    kw: dict[str, Any] = field(factory=dict)

    origin = _gen(0)
    timestamp = _gen(1)

    source: Any = None

    def __init__(self, /, name: str | EllipsisType | None = None, **kwargs):  # pyright: ignore
        vals = {}
        for k in dir(self):
            v = getattr(type(self), k, None)
            if isinstance(v, property) and (val := kwargs.pop(k, NotGiven)) is not NotGiven:
                vals[k] = val
        self.__attrs_init__()  # pyright:ignore
        self.kw = kwargs
        for k, v in vals.items():
            setattr(self, k, v)
        self._clean(name)

    def __repr__(self):
        res = f"‹{self.__class__.__name__} {self.a!r}"
        if self.kw:
            res += ","+repr(self.kw)
        if self.source is not None:
            res += " src="+self.source
        res += "›"
        return res

    def dump(self):
        """
        Emit this as an array, ready for sending.
        """
        if self.kw or self.a and isinstance(self.a[-1], dict):
            return self.a + [self.kw]
        return self.a

    def repr(self):
        """
        Emit this as a dict, ready for debugging.
        """
        kw = dict(self.kw)
        kw["origin"] = self.origin
        kw["timestamp"] = self.timestamp
        kw["_timestamp"] = ts2iso(self.timestamp, delta=True, msec=3)
        return kw

    @classmethod
    def restore(cls, a, kw=NotGiven):
        m = object.__new__(cls)
        if kw is NotGiven:
            if a and isinstance(a[-1], dict):
                kw = a.pop()
            else:
                kw = {}
        source = kw.pop("source", None)
        if isinstance(a, tuple):
            raise TypeError("why a tuple?")  # XXX
            a = list(a)
        m.a = a
        m.kw = kw
        m.source = source
        return m
    _moat__restore = restore

    def __getitem__(self, k):
        if isinstance(k, int):
            return self.a[k]
        else:
            return self.kw[k]

    def __getstate__(self):
        return self.a, self.kw

    def __setstate__(self, akw):
        self.a, self.kw, *_ = akw

    def __setitem__(self, k, v):
        if isinstance(k, slice):
            raise ValueError("Only use positive indices")  # noqa:TRY004
        if isinstance(k, int):
            if k < 0:
                raise KeyError(k)  # only positive indices
            elif k == 0 and (not isinstance(v, str) or v == "" or "/" in v or "\\" in v):
                raise ValueError("First item must be a string")
            self._len(k + 1)
            self.a[k] = v
        else:
            self.kw[k] = v

    def __delitem__(self, k):
        if isinstance(k, int):
            self.a[k] = None
        else:
            del self.kw[k]

    def _clean(self, name: str | EllipsisType | None):
        if name is NotGiven:
            return
        if name is None:
            if not self.origin:
                raise ValueError("You need to set a name")
        elif not self.origin:
            assert isinstance(name,str)
            if not name.startswith("via:"):
                name = f"via:{name}"
            self.origin = name
        if not self.timestamp:
            self.timestamp = time.time()

    def _len(self, n):
        while len(self.a) < n:
            self.a.append(None)

    def _map(self):
        res = self.a[:]
        while res and res[-1] is None:
            res.pop()
        if self.kw or isinstance(res[-1], dict):
            res.append(self.kw)
        return res

    def _unmap(self, data: list[Any]):
        if isinstance(data[-1], dict):
            self.kw = data.pop()
        self.a = data
        dit = iter(data)
        try:
            self.origin = next(dit)
            self.timestamp = next(dit)
        except StopIteration:
            pass

    def encode(self) -> str:
        """
        Encode this object to a string.

        Elements are either UTF-8 strings, introduced by ``/``, or
        some other data, introduced by ``\\``. Strings that include
        either of these characters are treated as "other data".

        Empty strings are encoded as zero-length "other data" elements.
        A value of ``None`` is encoded as an empty string.

        The first item is not marked explicitly.
        It must be a non-empty string.

        Other data are encoded to CBOR, then base85-encoded
        (btoa alphabet).

        The last element may be a dict with free-form content.
        """
        data = self._map()
        res = []
        for d in data:
            if isinstance(d, str) and d != "" and "/" not in d and "\\" not in d:
                if not d and not res:
                    raise ValueError("No empty origins")
                if res:
                    res.append("/")
                if d is not None:
                    res.append(d)
                continue

            if not res:
                raise ValueError("No non-string origins")
            res.append("\\")
            if d != "":
                res.append(b85encode(_codec.encode(d)).decode("utf-8"))
        return "".join(res)

    @classmethod
    def decode(cls, name: str, data: str) -> Self:
        """
        Decode a string to a `MsgMeta` object.

        Reverses the effect of `encode`.
        """
        res = cls(NotGiven)
        ddec = []

        encoded = False
        while data:
            next_enc: bool
            cc: int

            c1 = data.find("/")
            c2 = data.find("\\")
            if c1 == -1:
                cc = c2
                next_enc = True
            elif c2 == -1:
                cc = c1
                next_enc = False
            else:
                cc = min(c1, c2)
                next_enc = cc == c2

            if cc == -1:
                d, data = data, ""
            else:
                d, data = data[:cc], data[cc + 1 :]
            if encoded:
                if d != "":
                    d = _codec.decode(b85decode(d.encode("utf-8")))
            elif d == "":
                d = None
            ddec.append(d)
            encoded = next_enc

        res._unmap(ddec)  # noqa:SLF001
        res._clean(name)  # noqa:SLF001
        return res

def _meta_repr(dumper, data):
    return dumper.represent_scalar("!Meta", repr(data))

SafeRepresenter.add_representer(MsgMeta, _meta_repr)




# command line interface
from __future__ import annotations

import datetime
import time
import anyio

import asyncclick as click
from moat.util import MsgReader, NotGiven, P, PathLongener, attr_args, yprint, Path
from moat.util.times import ts2iso, humandelta

from moat.link.client import Link
from moat.link._data import data_get, node_attr
from moat.link.meta import MsgMeta
from moat.link.node import Node


@click.group(short_help="Manage data.", invoke_without_command=True)  # pylint: disable=undefined-variable
@click.option("-m","--meta",is_flag=True,help="include metadata")
@click.argument("path", type=P, nargs=1)
@click.pass_context
async def cli(ctx, path,meta):
    """
    This subcommand accesses the data stored in the MoaT-Link server.
    """
    obj = ctx.obj
    cfg = obj.cfg["link"]
    obj.conn = await ctx.with_async_resource(Link(cfg, name=obj.name))
    obj.meta=meta
    if ctx.invoked_subcommand is None:
        res = await data_get(obj.conn,path, meta=obj.meta,out=obj.stdout, recursive=False)
    else:
        obj.path = path



@cli.command()
@click.option(
    "-d",
    "--as-dict",
    default=None,
    help="Structure as dictionary. The argument is the key to use "
    "for values. Default: return as list",
)
@click.option(
    "-m",
    "--maxdepth",
    type=int,
    default=None,
    help="Limit recursion depth. Default: whole tree",
)
@click.option(
    "-M",
    "--mindepth",
    type=int,
    default=None,
    help="Starting depth. Default: whole tree",
)
@click.option("-r", "--recursive", is_flag=True, help="Read a complete subtree")
@click.option("-e", "--empty", is_flag=True, help="Include empty nodes")
@click.option("-R", "--raw", is_flag=True, help="Print string values without quotes etc.")
@click.option("-D", "--add-date", is_flag=True, help="Add *_date entries")
@click.pass_obj
async def get(obj, **k):
    """
    Read a MoaT-KV value.

    If you read a sub-tree recursively, be aware that the whole subtree
    will be read before anything is printed. Use the "watch --state" subcommand
    for incremental output.
    """

    await data_get(obj.conn, obj.path, meta=obj.meta, **k)


@cli.command("list")
@click.option(
    "-d",
    "--as-dict",
    default=None,
    help="Structure as dictionary. The argument is the key to use "
    "for values. Default: return as list",
)
@click.option(
    "-m",
    "--maxdepth",
    type=int,
    default=1,
    help="Limit recursion depth. Default: 1 (single layer).",
)
@click.option(
    "-M",
    "--mindepth",
    type=int,
    default=1,
    help="Starting depth. Default: 1 (single layer).",
)
@click.pass_obj
async def list_(obj, **k):
    """
    List MoaT-KV values.

    This is like "get" but with "--mindepth=1 --maxdepth=1 --recursive --empty"

    If you read a sub-tree recursively, be aware that the whole subtree
    will be read before anything is printed. Use the "watch --state" subcommand
    for incremental output.
    """

    k["recursive"] = True
    k["raw"] = True
    k["empty"] = True
    await data_get(obj.conn, obj.path, meta=obj.meta, **k)


@cli.command("set", short_help="Add or update an entry")
@attr_args
@click.option("-l", "--last", nargs=2, help="Previous change entry (node serial)")
@click.option("-n", "--new", is_flag=True, help="This is a new entry.")
@click.pass_obj
async def set_(obj, last, new, **kw):
    """
    Store a value at some MoaT-KV position.

    If you update a value you can use "--last" to ensure that no other
    change arrived between reading and writing the entry.

    When adding a new entry use "--new" to ensure that you don't
    accidentally overwrite something.

    MoaT-KV entries typically are mappings. Use a colon as the path if you
    want to replace the top level.
    """

    res = await node_attr(obj, obj.path, **kw)

    if obj.meta:
        yprint(res, stream=obj.stdout)


class nstr:
    def __new__(cls, val):
        if val is NotGiven:
            return val
        return str(val)


@cli.command(short_help="Delete an entry / subtree")
@click.option(
    "-b",
    "--before",
    type=float,
    help="Don't delete entries created after this timestamp",
)
@click.option("-r", "--recursive", is_flag=True, help="Delete a complete subtree")
@click.pass_obj
async def delete(obj, before, recursive):
    """
    Delete an entry, or a subtree.

    You really should use "--before" flag to ensure that no other change
    arrived after you viewed the data in question.

    Non-recursively deleting an entry with children works and does *not*
    affect the child entries.

    The root entry cannot be deleted.
    """
    args = {}
    if recursive:
        args["rec"]=recursive
    if before:
        args["ts"]=before
    res = await obj.conn.d.delete(obj.path, **args)
    if obj.meta:
        res=dict(data=res[0],meta=MsgMeta.restore(res[1:]).repr())
    else:
        res = res[0]
    yprint(res, stream=obj.stdout)


@cli.command()
@click.option("-s", "--state", is_flag=True, help="Also get the current state.")
@click.option("-o", "--only", is_flag=True, help="Value only, nothing fancy.")
@click.option("-p", "--path-only", is_flag=True, help="Value only, nothing fancy.")
@click.option("-D", "--add-date", is_flag=True, help="Add *_date entries")
@click.option("-i", "--ignore", multiple=True, type=P, help="Skip this (sub)tree")
@click.pass_obj
async def monitor(obj, state, only, path_only, add_date, ignore):
    """Monitor a MoaT-Link subtree"""

    cfg = obj.cfg.link
    seen = False
    data = Node()
    plen=len(obj.path)+1


    def pr(res):
        if only:
            res = res.get("data", NotGiven)
        elif path_only:
            res = res["path"]
        elif not obj.meta:
            del res["meta"]
        yprint(res, stream=obj.stdout)
        print("---", file=obj.stdout)
        obj.stdout.flush()

    def set(p,d,m):
        r = dict(path=p,meta=m.repr())
        if d is not NotGiven:
            r["data"] = d
        dd = data.get(p)
        ddm = dd.meta
        n = dd.set(..., d, m)
        if ddm is not None:
            r["last"] = ts = ddm.timestamp-time.time()
            r["_last"] = humandelta(ts, ago=True)
        if n is False:
            r["_***_"]="Obsolete message. Ignored."
        return r

    async with (
        anyio.create_task_group() as tg,
        obj.conn.monitor(cfg.root+obj.path, subtree=True) as res,
    ):
        if state:
            @tg.start_soon
            async def get_tree():
                pl = PathLongener(())
                async with obj.conn.d.walk(obj.path) as mon:
                    async for n,p,d,*m in mon:
                        p = pl.long(n,p)
                        m = MsgMeta.restore(m)
                        res = set(p,d,m)
                        pr(res)

        async for msg in res:
            mp = Path.build(msg.topic[plen:])
            if any(p == mp[: len(p)] for p in ignore):
                continue

            res = set(mp,msg.data,msg.meta)
            pr(res)


@cli.command()
@click.option("-i", "--infile", type=click.Path(), help="File to read.")
@click.option("-C", "--codec", type=str, default="yaml", help="Codec to use (default: yaml).")
@click.pass_obj
async def update(obj, infile, codec):
    """Write a list of updates to a MoaT-Link subtree"""
    async with MsgReader(path="/dev/stdin" if infile == "-" else infile, codec=codec) as reader:
        async for msg in reader:
            if isinstance(msg,dict):
                p = msg["path"]
                v = msg["value"]
            else:
                p,v,*_m = msg
            await obj.conn.d_set(obj.path + p, v)



"""
Common code for backends.

Currently the only backend supported is MQTT, but who knows.
"""

from __future__ import annotations

import anyio
import logging
from abc import ABCMeta, abstractmethod
from contextlib import asynccontextmanager

from attrs import define, field

from moat.util import CtxObj, NotGiven, Path, Root, RootPath, attrdict, get_codec

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.codec import Codec
    from moat.link.meta import MsgMeta

    from typing import Any, Self, ClassVar
    from collections.abc import AsyncIterator


__all__ = ["get_backend", "get_codec", "Backend", "Message", "RawMessage"]


@define
class Message[TData]:
    """
    An incoming message.
    """

    topic: Path = field()
    data: TData = field()
    meta: MsgMeta = field()
    prop: dict[str,Any] = field()
    retain: bool|None = field(default=False)

    raw: ClassVar[bool] = False

    def __class_getitem__(cls, TData):
        return cls  # for now


@define
class RawMessage(Message):
    "A message that couldn't be decoded / shouldn't be encoded"

    exc: Exception = field(default=None)
    raw: ClassVar[bool] = True


class Backend(CtxObj, metaclass=ABCMeta):
    "Base class for messaging backends"

    _tg: anyio.abc.TaskGroup
    _njobs: int = 0
    _ended: anyio.Event | None = None

    def __init__(self, cfg: attrdict, name: str, id:str):
        self.cfg = cfg
        self.name = name
        self.id = id
        self.logger = logging.getLogger(f"moat.link.backend.{name}")

    @abstractmethod
    @asynccontextmanager
    async def connect(self, *a, **k) -> AsyncIterator[Self]:
        """
        This async context manager returns a connection to the backend.
        """

    @asynccontextmanager
    async def _ctx(self) -> AsyncIterator[Self]:
        async with (
            anyio.create_task_group() as self._tg,
            self.connect(),
        ):
            yield self

    @abstractmethod
    @asynccontextmanager
    async def monitor(
        self,
        topic: Path,
        qos: QoS | None = None,
        codec: Codec | None | Literal[NotGiven] = NotGiven,
        raw: bool | None = False,
        echo: bool = False,
        no_local: bool = False,
        subtree: bool = False,
    ) -> AsyncIterator[AsyncIterator[Message]]:
        """
        Return an async iterator that listens to this topic.

        Set @raw to ``True`` to always get non-decoded messages.
        """

    @abstractmethod
    async def send(
        self,
        topic: Path,
        data: Any,
        codec: Codec | None | Literal[NotGiven] = NotGiven,
        **kw: Any,
    ) -> None:
        """
        Send this payload to this topic.
        """

    async def send_error(
        self,
        subpath: Path,
        msg: str | None = None,
        data: Any = None,
        exc: Exception | None = None,
        **kw: Any,
    ):
        """
        Send a somewhat-free-form error message.
        """
        if not isinstance(subpath[0], RootPath):
            subpath = Root.get() / "error" + subpath
        if msg:
            kw["msg"] = msg
        if data is not NotGiven:
            kw["data"] = data
        if exc is not None:
            kw["exc"] = repr(exc)
        # TODO include the backtrace?

        try:
            await self.send(subpath, kw, codec="std-cbor")
        except Exception:
            self.logger.exception("Failure to log error at %s: %r", subpath, data, exc_info=exc)
        else:
            self.logger.debug("Error at %s: %r", subpath, data, exc_info=exc)


def get_backend(cfg: dict, **kw) -> Backend:
    """
    Fetch the backend named in the config and initialize it.
    """
    from importlib import import_module

    cfg = cfg["backend"]
    name = cfg["driver"]
    if "." not in name:
        name = "moat.link.backend." + name
    return import_module(name).Backend(cfg, **kw)



"""
Data access
"""

from __future__ import annotations

import anyio
import datetime
import os
import sys
import time
from collections.abc import Mapping

from .node import Node
from .meta import MsgMeta
from moat.util import NotGiven, Path, attrdict, process_args, yprint, PathLongener
from moat.util.times import ts2iso
import contextlib


def add_dates(d):
    """
    Given a dict with int/float entries that might conceivably be dates,
    add ``_*`` with a textual representation.
    """

    t = time.time()
    start = t - 366 * 24 * 3600
    stop = t + 366 * 24 * 3600

    def _add(d):
        if isinstance(d, (list, tuple)):
            for dd in d:
                _add(dd)
            return
        if not isinstance(d, Mapping):
            return
        for k, v in list(d.items()):
            if isinstance(k, str) and k.startswith("_"):
                continue
            if not isinstance(v, (int, float)):
                _add(v)
                continue
            if start <= v <= stop:
                d[f"_{k}"] = ts2iso(v)

    _add(d)


async def data_get(
    conn:Link,
    path:Path,
    *,
    meta:bool=False,
    recursive:bool=True,
    as_dict:str|None="_",
    maxdepth:int=-1,
    mindepth:int=0,
    empty:bool=False,
    raw:bool=False,
    path_mangle=None,
    item_mangle=None,
    add_date=False,
    out=None,
):
    """Generic code to dump a subtree.

    `path_mangle` accepts a path and the as_dict parameter. It should
    return the new path. This is used for e.g. prefixing the path with a
    device name. Returning ``None`` causes the entry to be skipped.
    """
    if path_mangle is None:

        def path_mangle(x):
            return x

    if item_mangle is None:

        async def item_mangle(x):  # pylint: disable=function-redefined
            return x

    if out is None:
        out = sys.stdout
    elif out is False:
        out = []

    if recursive:
        kw = {}
        
        if maxdepth is not None and maxdepth >= 0:
            kw["max_depth"] = maxdepth
        if mindepth:
            kw["min_depth"] = mindepth
        if empty:
            kw["empty"] = True
        y = {}
        pl=PathLongener()
        async with conn.d.walk(path, **kw).stream_in() as res:
            async for r in res:
                r = await item_mangle(r)
                if r is None:
                    continue
                n,p,d,*m = r
                path=pl.long(n,p)
                path = path_mangle(path)
                if path is None:
                    continue

                if add_date:
                    add_dates(d)

                if meta:
                    m=MsgMeta._moat__restore(m, NotGiven)

                if as_dict is not None:
                    if meta:
                        d = dict(data=d,meta=m.repr())
                    yy = y
                    for p in path:
                        yy = yy.setdefault(p, {})
                    try:
                        yy[as_dict] = d
                    except AttributeError:
                        if empty:
                            yy[as_dict] = None
                else:
                    if raw:
                        y = path
                    elif meta:
                        y = [path,d,m.repr()]
                    else:
                        y = [path,d]
                    if isinstance(out,list):
                        out.append(y)
                    else:
                        yprint(y, stream=out)
                        out.write("---\n")

            if as_dict is not None:
                if maxdepth:

                    def simplex(d):
                        for k, v in d.items():
                            if isinstance(v, dict):
                                d[k] = simplex(d[k])
                        if as_dict in d and d[as_dict] is None:
                            if len(d) == 1:
                                return None
                            else:
                                del d[as_dict]
                        return d

                    y = simplex(y)

            if isinstance(out,list):
                return y
            if as_dict is not None:
                yprint(y, stream=out)

            return out # end "if recursive"

    
    d,*m = await conn.d.get(path)
    if add_date:
        add_dates(d)
    if meta:
        m=MsgMeta.restore(m)
        d=dict(data=d,meta=m.repr())

    if out is False:
        return d
    if not raw:
        yprint(d, stream=out)
    elif isinstance(d, bytes):
        os.write(out.fileno(), res)
    else:
        out.write(str(d))
    pass  # end get


async def backend_get(
    conn:Link,
    path:Path,
    *,
    meta:bool=False,
    recursive:bool=True,
    as_dict:str|None="_",
    raw:bool=False,
    path_mangle=None,
    item_mangle=None,
    add_date=False,
    codec:Codec|None=None,
    out=None,
    timeout:float=0.5,
):
    """Generic code to dump a backend subtree.

    `path_mangle` accepts a path and the as_dict parameter. It should
    return the new path. This is used for e.g. prefixing the path with a
    device name. Returning ``None`` causes the entry to be skipped.
    """
    # This is a copy of `data_get` that accesses the backend directly.

    if path_mangle is None:

        def path_mangle(x):
            return x

    if item_mangle is None:

        async def item_mangle(x):  # pylint: disable=function-redefined
            return x

    if out is None:
        out = sys.stdout
    elif out is False:
        out = []

    if recursive:
        kw = {"codec":codec}
        
        y = {}
        async with conn.monitor(path, subtree=True, **kw) as mon:
            while True:
                r = None
                with anyio.move_on_after(timeout):
                    r = await anext(mon)
                if r is None:
                    break

                r = await item_mangle(r)
                if r is None:
                    continue
                p,d,m = r.topic,r.data,r.meta
                p=p[len(path):]
                p = path_mangle(p)
                if p is None:
                    continue

                if add_date:
                    add_dates(d)
                if meta:
                    m=MsgMeta._moat__restore(m, NotGiven)
                    d = dict(data=d,meta=m.repr())

                if as_dict is not None:
                    yy = y
                    for pp in p:
                        yy = yy.setdefault(pp, {})
                    try:
                        yy[as_dict] = d
                    except AttributeError:
                        if empty:
                            yy[as_dict] = None
                else:
                    if raw:
                        y = p
                    else:
                        y = {}
                        try:
                            y[p] = d
                        except AttributeError:
                            if empty:
                                y[p] = None
                            else:
                                continue
                    if isinstance(out,list):
                        out.append(y)
                    else:
                        yprint([y], stream=out)

            if isinstance(out,list):
                return y
            yprint(y, stream=out)

            return out # end "if recursive"

    
    async with conn.monitor(path, **kw) as mon:
        r = None
        with anyio.move_on_after(timeout):
            r = await anext(mon)
        if r is None:
            raise KeyError(path)

    d,m = r.data,r.meta
    if add_date:
        add_dates(d)
    if meta:
        m=MsgMeta.restore(m)
        d=dict(data=d,meta=m.repr())

    if out is False:
        return d
    if not raw:
        yprint(d, stream=out)
    elif isinstance(d, bytes):
        os.write(out.fileno(), res)
    else:
        out.write(str(d))
    pass  # end get


def res_get(res, attr: Path, **kw):  # pylint: disable=redefined-outer-name
    """
    Get a node's value and access the dict items beneath it.

    The node value must be an attrdict.
    """
    val = res.get("value", None)
    if val is None:
        return None
    return val._get(attr, **kw)


def res_update(res, attr: Path, value=None, **kw):  # pylint: disable=redefined-outer-name
    """
    Set a node's sub-item's value, possibly merging dicts.
    Entries set to 'NotGiven' are deleted.

    The node value must be an attrdict.

    Returns the new value.
    """
    val = res.get("value", attrdict())
    return val._update(attr, value=value, **kw)


async def node_attr(obj, path, val=NotGiven, meta=NotGiven, **kw):
    """
    Sub-attr setter.

    Args:
        obj: command object
        path: address of the node to change
        res: old node, if it has been read already
        **kw: the results of `attr_args`

    Returns the result of setting the attribute.
    """
    if val is NotGiven:
        try:
            val,*m = await obj.conn.d.get(path)
        except KeyError:
            pass
        else:
            meta = MsgMeta.restore(m)
    val = process_args(val, **kw)
    t = {} if meta is NotGiven else {'t':meta.timestamp}
    if val is NotGiven:
        res = await obj.conn.d.delete(path, **t)
    else:
        res = await obj.conn.d.set(path, val, **t)
    res = res[0],MsgMeta.restore(res[1:])
    return res



"""
This module contains all MoaT-Link exceptions.

Currently this is a re-import of moat.kv.exceptions.
"""

from __future__ import annotations

from moat.kv.exceptions import *

MoaTLinkError = MoaTKVError
